/*
 * directory.c
 *
 * Copyright (c) 1999  Oliver Fromme  <olli@fromme.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *     @(#)$Id: directory.c,v 1.3 1999/09/02 21:12:09 olli Exp $
 */

static const char cvsid[]
    = "@(#)$Id: directory.c,v 1.3 1999/09/02 21:12:09 olli Exp $";

#include <string.h>
#include <time.h>
#include <ctype.h>

#include "utils.h"

#include "ophoto.h"
#include "directory.h"

/*
 *   Note:  The camera itself uses backslashes as path
 *   separators.  However, we use forward slashes in all
 *   upper-level functions, se we stay compatible with
 *   UNIX conventions.  Conversion to backslashes is
 *   handled in the lower-level functions for the camera
 *   only.
 */

directory_entry root_entry = {".", ATTR_DIR, 0, 0, NULL};

/*
 *   Parse a file name in "FAT" format and convert
 *   to standard UNIX format.
 */

char *
parse_fatname (char *dst, char *fatname)
{
	int j, k;

	for (j = 0; j < 8 && fatname[j] != ' '; j++)
		dst[j] = tolower(fatname[j]);
	if (fatname[8] && fatname[8] != ' ') {
		dst[j++] = '.';
		for (k = 0; k < 3 && fatname[8 + k] != ' '; k++)
			dst[j + k] = tolower(fatname[8 + k]);
		dst[j + k] = 0;
	}
	else
		dst[j] = 0;
	return dst;
}

/*
 *   Canonicalize a path:
 *   Convert all slashes to forward slashes, compress sequences
 *   of slashes, make sure that the path begins with a slash
 *   and does not end with one.  Excpetion:  The root directory
 *   is the empty string.
 *
 *   If the given name is relative (i.e. does not start with
 *   a slash), we prepend "/dcim/", just as if this was the
 *   current working directory.
 */

char *
canon_path (char *src)
{
	char *new, *dst;
	bool have_slash;

	if (*src == '\\' || *src == '/')
		new = dst = tmalloc(strlen(src) + 1);
	else {
		new = tmalloc(strlen(src) + 7);
		strcpy (new, "/dcim");
		dst = strchr(new, 0);
	}
	*dst++ = '/';
	have_slash = TRUE;
	for (;;) {
		while (*src == '\\' || *src == '/')
			if (have_slash)
				src++;
			else {
				*dst++ = '/';
				have_slash = TRUE;
			}
		if (!*src)
			break;
		*dst++ = tolower(*src++);
		have_slash = FALSE;
	}
	if (have_slash)
		dst--;
	*dst = 0;
	return new;
}

static time_t
mktimet (int year, int mon, int day, int hour, int min, int sec)
{
	struct tm stm;

	stm.tm_sec = sec;
	stm.tm_min = min;
	stm.tm_hour = hour;
	stm.tm_mday = day;
	stm.tm_mon = mon - 1;
	stm.tm_year = year - 1900;
	stm.tm_isdst = -1;
	return mktime(&stm);
}

/*
 *   Fetches the given directory.
 *   The path name must be canonic, and it must exist.
 *   This function internally converts slashes to backslashes
 *   for the camera.
 */

static directory *
fetch_dir (kodacon *kc, char *dirname)
{
	directory *dir;
	directory_entry *entry;
	char *camname, *cptr;
	int result, i, crdate, crtime;
	byte *data;

	dir = tmalloc(sizeof(directory));
	camname = tmalloc(strlen(dirname) + 1);
	cptr = camname;
	while (*dirname)
		if (*dirname == '/') {
			dirname++;
			*cptr++ = '\\';
		}
		else
			*cptr++ = *dirname++;
	*cptr = 0;
	if (kc->fd_in < 0)
		init_connection (kc);
	result = read_directory(kc, camname);
	if (kc->verbosity > 1 || result < 0)
		print_debug (kc, result, "reading directory");	
	free (camname);
	dir->num_entries = kc->data[0] << 8 | kc->data[1];
	dir->entries = tmalloc(dir->num_entries * sizeof(directory_entry));
	for (i = 0; i < dir->num_entries; i++) {
		data = kc->data + i * 20 + 2;
		entry = dir->entries + i;
		parse_fatname (entry->name, (char *) data);
		entry->attr = data[11];
		crtime = data[12] << 8 | data[13];
		crdate = data[14] << 8 | data[15];
		entry->datetime = mktimet((crdate >> 9) + 1980,
		    (crdate >> 5 & 15) + 1, crdate & 31,
		    crtime >> 11, crtime >> 5 & 63, (crtime & 31) * 2);
		entry->size = big2ul(data + 16);
		entry->subdir = NULL;
	}
	return dir;
}

/*
 *   Check if a directory entry exists and has the correct type
 *   (dir_type > 0 indicates that it should be a directory, ==
 *   0 indicates that it should be a plain file, and < 0 means
 *   that we don't care).  Makes sure that all directories are
 *   fetched if necessary.
 *
 *   Returns a pointer to the corresponding directory entry.
 *   Returns NULL if the entry does not exist or is not of the
 *   requested type.
 *
 *   The path is expected to be canonic, see canon_path().
 */

directory_entry *
check_dir_entry (kodacon *kc, char *path, int dir_type)
{
	char *rslash, *dirname;
	directory *parent;
	directory_entry *entry;
	int i;
	bool is_dir;

	/*
	 *   The root directory always exists (and it's
	 *   always a directory, obviously).
	 */

	if (!*path) {
		if (dir_type == 0)
			return NULL;
		if (!root_entry.subdir) {
			root_entry.subdir = fetch_dir(kc, "");
			if (root_entry.subdir->num_entries > 0) {
				/*
				 *   We copy the date/time of the "." entry
				 *   of the root directory to our dummy root
				 *   entry.  We assume that "." is the first
				 *   directory entry.
				 */
				root_entry.datetime =
				    root_entry.subdir->entries->datetime;
			}
			else
				root_entry.datetime = 0;
		}
		return &root_entry;
	}

	/*
	 *   Find the rightmost slash and then check if
	 *   the path exists up to that point.
	 */

	if (!(rslash = strrchr(path, '/')))
		return NULL;
	if (!(dirname = strndup(path, rslash - path)))
		out_of_memory();
	entry = check_dir_entry(kc, dirname, 1);
	free (dirname);
	if (!entry || (entry->attr & ATTR_DIR) == 0)
		return NULL;
	parent = entry->subdir;
	rslash++;

	/*
	 *   Now search the entry in the directory.
	 *   Everything has already been converted to
	 *   lowercase, so we needn't care about that.
	 */

	for (i = 0; i < parent->num_entries; i++) {
		if (strcmp(parent->entries[i].name, rslash) == 0)
			break;
	}
	if (i >= parent->num_entries)
		return NULL;
	entry = parent->entries + i;
	is_dir = (entry->attr & ATTR_DIR) != 0;
	if ((dir_type > 0 && !is_dir) || (dir_type == 0 && is_dir))
		return NULL;
	if (is_dir && !entry->subdir)
		entry->subdir = fetch_dir(kc, path);
	return entry;
}

/* EOF */
