/*
 * ophoto.c
 *
 * Copyright (c) 1999  Oliver Fromme  <olli@fromme.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *     @(#)$Id: ophoto.c,v 1.8 1999/12/19 07:13:59 olli Exp $
 */

static const char cvsid[]
    = "@(#)$Id: ophoto.c,v 1.8 1999/12/19 07:13:59 olli Exp $";

#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <sysexits.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <err.h>

#include </sys/dev/usb/usb.h>

#include "utils.h"
#include "getlopt.h"

#include "ophoto.h"
#include "command.h"
#include "proto.h"

#define DEFAULT_DEVICE	"/dev/ugen0"
#define DEFAULT_EP_OUT	1
#define DEFAULT_EP_IN	2

char *device = DEFAULT_DEVICE;	/* device base name */
int ep_out = DEFAULT_EP_OUT;	/* number of output endpoint */
int ep_in  = DEFAULT_EP_IN;	/* number of input endpoint */
int verbosity = 0;		/* verbosity level */
int debuglevel = 0;		/* debug level */

static void
incr_verb (char *arg)
{
	verbosity++;
}

static void
decr_verb (char *arg)
{
	verbosity--;
}

static void
incr_debug (char *arg)
{
	debuglevel++;
}

static void
set_device (char *dev)
{
	if (strchr(dev, '/')) {
		/*
		 *   Full path to device specified, e.g. "/dev/ugen0".
		 */
		device = strdup(dev);
	}
	else {
		if (*dev >= '0' && *dev <= '9') {
			/*
			 *   Only device number specified, e.g. "0".
			 */
			device = malloc(strlen(dev) + 10);
			strcpy (device, "/dev/ugen");
			strcat (device, dev);
		}
		else {
			/*
			 *   Only device name specified, e.g. "ugen0".
			 */
			device = malloc(strlen(dev) + 6);
			strcpy (device, "/dev/");
			strcat (device, dev);
		}
	}
}

static void
set_endpoint (int *ep, char *s)
{
	char *eptr;

	if (s && *s) {
		*ep = strtol(s, &eptr, 10);
		if (!*eptr || *ep < 1 || *ep > 15)
			errx (EX_USAGE, "illegal endpoint number \"%s\"", s);
	}
}

static void
set_endpoints (char *dev)
{
	char *cptr;

	if ((cptr = strchr(dev, ','))) {
		/*
		 *   If two endpoint numbers are specified
		 *   (separated by a comma, e.g. "1,2"),
		 *   then set both accordingly.
		 */
		set_endpoint (&ep_out, strndup(dev, cptr - dev));
		set_endpoint (&ep_in, strdup(cptr + 1));
	}
	else {
		/*
		 *   If only one endpoint number is specified,
		 *   then set the output endpoint to be this number,
		 *   and set the input endpint to be number + 1.
		 */
		set_endpoint (&ep_out, strndup(dev, cptr - dev));
		ep_in = ep_out + 1;
	}
}

static void
usage (char *arg)
{
	fprintf (stderr, "oPhoto ALPHA version, 19-Dec-1999\n");
	fprintf (stderr, "Usage:  %s [<options>] <command> ...\n", me);
	fprintf (stderr, "Standard options:\n");
	fprintf (stderr, "   -d <device>     device (default: %s)\n", DEFAULT_DEVICE);
	fprintf (stderr, "   -v              be more verbose\n");
	fprintf (stderr, "   -q              be less verbose (quiet)\n");
	fprintf (stderr, "Wizard options (do not use if you don't know exactly what they mean!):\n");
	fprintf (stderr, "   -e <out>,<in>   endpoints (default: %d,%d)\n", DEFAULT_EP_OUT, DEFAULT_EP_IN);
	fprintf (stderr, "   -D              increase debug level\n");
	fprintf (stderr, "Type \"%s help\" for help on commands.\n", me);
	exit (EX_USAGE);
}

const topt myopts[] = {
	{'d', "device",    GLO_CHAR, set_device,    0, 0},
	{'v', "verbose",   GLO_NONE, incr_verb,     0, 0},
	{'q', "quiet",     GLO_NONE, decr_verb,     0, 0},
	{'e', "endpoint",  GLO_CHAR, set_endpoints, 0, 0},
	{'D', "debug",     GLO_NONE, incr_debug,    0, 0},
	{'h', "help",      GLO_NONE, usage,         0, 0},
	{'?', 0, GLO_NONE, usage, 0, 0},
	{0, 0, 0, 0, 0, 0}
};

void
print_debug (kodacon *kc, int result, char *what)
{
	switch (result) {
	case 0:
		warnx ("%s: OK", what);
		break;
	case -1:
		err (EX_UNAVAILABLE, "%s", what);
	case -2:
		errx (EX_PROTOCOL, "%s: %s", what, strsyscode(kc->syscode));
	default:
		errx (EX_SOFTWARE, "%s: unknown error!", what);
	}
}

/*
 *   Save a file to the disk.  The name "-" means stdout.
 *   Partial writes are handled correctly;
 *
 *   Returns 0 on success, -1 on error (with errno set).
 */

int
save_file (char *name, byte *data, size_t size)
{
	int fd, result, is_stdout;

	result = 0;
	is_stdout = !strcmp(name, "-");
	if (is_stdout)
		fd = fileno(stdout);
	else if ((fd = open(name, O_CREAT | O_TRUNC | O_WRONLY, 0644)) < 0)
		return -1;
	while (size) {
		while ((result = write(fd, data, size)) < 0 && errno == EINTR)
			;
		if (result < 0)
			break;
		data += result;
		size -= result;
	}
	if (!is_stdout) {
		/* close() might overwrite errno, so we have to save it. */
		int errtmp = errno;
		while (close(fd) < 0 && errno == EINTR)
			;
		errno = errtmp;
	}
	return result;
}

/*
 *   Set up device names and open the endpoints.
 *   The Kodak DC240 uses a very simple protocol via
 *   two bulk endpoints (input and output).  It only
 *   has one configuration and one interface, so we
 *   don't have to care about these.
 */

void
init_connection (kodacon *kc)
{
	char *dev_out, *dev_in;	/* USB device node names */
	int ufd_out, ufd_in;	/* USB device file descriptors */
	int result;

	asprintf (&dev_out, "%s.%d", device, ep_out);
	asprintf (&dev_in, "%s.%d", device, ep_in);
	if (verbosity > 0)
		fprintf (stderr, "%s: using device %s\n", me, device);
	if (verbosity > 1) {
		fprintf (stderr, "%s: using endpoint #%d (%s) for output\n",
		    me, ep_out, dev_out);
		fprintf (stderr, "%s: using endpoint #%d (%s) for input\n",
		    me, ep_in, dev_in);
	}
	if ((ufd_out = open(dev_out, O_WRONLY, 0)) < 0)
		err (EX_CONFIG, "open(\"%s\",O_WRONLY)", dev_out);
	if ((ufd_in = open(dev_in, O_RDONLY, 0)) < 0)
		err (EX_CONFIG, "open(\"%s\",O_RDONLY)", dev_in);
	result = 1;
	if (ioctl(ufd_in, USB_SET_SHORT_XFER, &result) < 0)
		err (EX_OSERR, "ioctl(``%s'',USB_SET_SHORT_XFER)", dev_in);
	result = proto_init(kc, ufd_out, ufd_in, verbosity, debuglevel);
	if (verbosity > 1 || result < 0)
		print_debug (kc, result, "initializing protocol");
	free (dev_out);
	free (dev_in);
}

void
done_connection (kodacon *kc)
{
	int result;

	result = proto_done(kc);
	if (verbosity > 1 || result < 0)
		print_debug (kc, result, "finalizing protocol");
	close (kc->fd_in);
	close (kc->fd_out);
	kc->fd_in = kc->fd_out = -1;
}

int
main (int argc, char *argv[])
{
	kodacon *kc;		/* camera connection state information */
	int optinc;

	utils_init (argv[0]);
	parselopts (argc, argv, myopts, me);
	if (loptind >= argc)
		usage (NULL);
	kc = malloc(sizeof(kodacon));
	kc->fd_in = kc->fd_out = -1;
	while (loptind < argc) {
		optinc = execute_command(kc, argc - loptind, argv + loptind);
		loptind += optinc;
	}
	if (kc->fd_in >= 0)
		done_connection (kc);
	free (kc);
	return EX_OK;
}

/* EOF */
