/*
 * proto.c
 *
 * Copyright (c) 1999  Oliver Fromme  <olli@fromme.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#ifdef ENABLE_TIMEOUT
#include <signal.h>
#include <sys/time.h>
#endif
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>
#include <errno.h>

#include "utils.h"

#include "proto.h"
#include "cmds.h"

#ifdef ENABLE_TIMEOUT
#define USB_READ_TIMEOUT 3	/* 3 seconds */
#endif

/*
 *   Generic notes:
 *
 *   Most of the functions (unless otherwise stated) return
 *   some positive result or 0 in case of success.
 *
 *   -1 is returned if an I/O error occured, and errno is
 *   set to the error code in this case (see intro(2)).
 *
 *   -2 is returned if the camera returned a system code
 *   indicating an error condition, such as a protocol error
 *   or checksum mismatch.  The system code is stored in
 *   kc->syscode in this case.
 */

#ifdef ENABLE_TIMEOUT

/*
 *   Signal handler and related stuff for timeout functions.
 */

volatile sig_atomic_t gotalarm;

void
alarmhandler (int sig)
{
	gotalarm = 1;
}

void
init_alarmhandler (void)
{
	struct itimerval mytimer;

	signal (SIGALRM, alarmhandler);
	timerclear(&mytimer.it_interval);
	mytimer.it_value.tv_sec  = USB_READ_TIMEOUT;
	mytimer.it_value.tv_usec = 0;
	setitimer (ITIMER_REAL, &mytimer, NULL);
	gotalarm = 0;
}

void
done_alarmhandler (void)
{
	struct itimerval mytimer;

	timerclear(&mytimer.it_value);
	setitimer (ITIMER_REAL, &mytimer, NULL);
}

#endif

/*
 *   Calculate and verify the checksum of a data packet.
 *   The last byte of the data should contain the checksum
 *   to be verified.  The data should NOT contain the
 *   packet control byte (PCB).
 *
 *   Returns 0 if the checksum is correct, -1 otherwise.
 */

static int
verify_checksum (byte *data, int len)
{
	int i;
	byte cksum;

	len--;
	cksum = 0;
	for (i = 0; i < len; i++)
		cksum ^= data[i];
	return (cksum == data[i]) ? 0 : -1;
}

/*
 *   Create the checksum of a data packet.
 *   It will be written to the last byte of the data.
 *   The data should NOT contain the packet control byte.
 */

static void
create_checksum (byte *data, int len)
{
	int i;
	byte cksum;

	len--;
	cksum = 0;
	for (i = 0; i < len; i++)
		cksum ^= data[i];
	data[i] = cksum;
}

/*
 *   Send a system code to the camera.
 *   Returns 0 on success and -1 on I/O error (with errno set).
 */

static int
send_syscode (kodacon *kc, byte syscode)
{
	int result;

	if (kc->debuglevel > 1)
		fprintf (stderr, "<-- %02x\n", syscode);
	for (;;) {
		while ((result = write(kc->fd_out, &syscode, 1)) < 0
		    && errno == EINTR)
			;
		if (result < 0) {
			if (kc->debuglevel > 0)
				fprintf (stderr, "Error: %s\n",
				    strerror(errno));
			return -1;
		}
		if (result < 1) {
			if (kc->debuglevel > 2)
				fprintf (stderr, "write() returned 0\n");
			continue;
		}
		if (kc->debuglevel > 2)
			fprintf (stderr, "write() successful\n");
		return 0;
	}
}

/*
 *   Read a system code from the camera.
 *   Returns -1 on I/O error (with errno set).
 */

static int
read_syscode (kodacon *kc)
{
	int result;
	byte syscode;

	if (kc->debuglevel > 2)
		fprintf (stderr, "reading system code...\n");
	for (;;) {
#ifdef ENABLE_TIMEOUT
		do {
			init_alarmhandler();
			result = read(kc->fd_in, &syscode, 1);
		} while (result < 0 && errno == EINTR && !gotalarm);
#else
		do
			result = read(kc->fd_in, &syscode, 1);
		while (result < 0 && errno == EINTR);
#endif
		if (result < 0 && errno == EIO) {    /* Try again on EIO... */
			if (kc->verbosity >= 0)
				fprintf (stderr, "E");
#ifdef ENABLE_TIMEOUT
			do {
				init_alarmhandler();
				result = read(kc->fd_in, &syscode, 1);
			} while (result < 0 && errno == EINTR && !gotalarm);
#else
			do
				result = read(kc->fd_in, &syscode, 1);
			while (result < 0 && errno == EINTR);
#endif
		}
#ifdef ENABLE_TIMEOUT
		done_alarmhandler();
#endif
		if (result < 0) {
#ifdef ENABLE_TIMEOUT
			if (errno == EINTR && gotalarm) {
				if (kc->verbosity >= 0)
					fprintf (stderr, "\n");
				fprintf (stderr, "%s: USB/camera timeout\n",
				    me);
			}
#endif
			if (kc->debuglevel > 0)
				fprintf (stderr, "Error: %s\n",
				    strerror(errno));
			return -1;
		}
		if (result < 1) {
			if (kc->debuglevel > 2)
				fprintf (stderr, "write() returned 0\n");
			continue;
		}
		if (kc->verbosity >= 0)
			fprintf (stderr, ".");
		if (kc->debuglevel > 1)
			fprintf (stderr, "--> %02x\n", syscode);
		if (result == SYS_SUCCESS) {
			/*
			 *   The spec requires a delay of 50ms
			 *   after a command has been completed.
			 */
			while (usleep(50000) < 0 && errno == EINTR)
			 	;
		}
		return syscode;
	}
}

/*
 *   Wait for one or two specific system codes.  Specify the
 *   second one as -1 if only one code is expected.
 *
 *   Returns the received system code, or:
 *   -1   I/O error, errno is set.
 *   -2   Protocol or camera error, kc->syscode is set.
 */

static int
check_syscode (kodacon *kc, int syscode1, int syscode2)
{
	int result;

	for (;;) {
		if ((result = read_syscode(kc)) < 0)
			return -1;
		if (result == syscode1)
			return result;
		if (syscode2 >= 0 && result == syscode2)
			return result;
		if (result == SYS_BUSY)
			continue;
		/* Treat everything else as an error. */
		if (kc->debuglevel > 2)
			fprintf (stderr,
			    "Error: unexpected system code %02x\n", result);
		kc->syscode = result;
		return -2;
	}
}

/*
 *   Generic function to read arbitrary data from the camera.
 *
 *   Returns the number of bytes read on success,
 *   and returns -1 on I/O error (with errno set).
 */

static int
read_data (kodacon *kc, byte *data, int len)
{
	int result;

	if (kc->debuglevel > 2)
		fprintf (stderr, "reading %d bytes...\n", len);
#ifdef ENABLE_TIMEOUT
	do {
		init_alarmhandler();
		result = read(kc->fd_in, data, len);
	} while (result < 0 && errno == EINTR && !gotalarm);
#else
	do
		result = read(kc->fd_in, data, len);
	while (result < 0 && errno == EINTR);
#endif
	if (result < 0 && errno == EIO) {	/* Try again on EIO... */
		if (kc->verbosity >= 0)
			fprintf (stderr, "E");
#ifdef ENABLE_TIMEOUT
		do {
			init_alarmhandler();
			result = read(kc->fd_in, data, len);
		} while (result < 0 && errno == EINTR && !gotalarm);
#else
		do
			result = read(kc->fd_in, data, len);
		while (result < 0 && errno == EINTR);
#endif
	}
#ifdef ENABLE_TIMEOUT
	done_alarmhandler();
	if (result < 0 && errno == EINTR && gotalarm) {
		if (kc->verbosity >= 0)
			fprintf (stderr, "A");
		result = read_syscode(kc);
		if (result < 0)
			return result;
		*data = result;
		return 1;
	}
#endif
	if (result < 0) {
		if (kc->debuglevel > 0)
			fprintf (stderr, "Error: %s\n",
			    strerror(errno));
		return -1;
	}
	if (kc->verbosity >= 0)
		fprintf (stderr, ".");
	if (kc->debuglevel > 1)
		fprintf (stderr, "have read %d bytes\n", result);
	return result;
}

/*
 *   Read a single data packet from the camera (including
 *   PCB and checksum) and tell it whether we have received
 *   it correctly.  Short reads are possible.
 *
 *   Return values:
 *   n    Success (n = number of bytes read, n >= 0).
 *   -1   I/O error, errno is set.
 *   -2   Protocol or camera error, kc->syscode is set.
 */

static int
read_packet (kodacon *kc, byte *data, int len)
{
	int result, num_retry;

	if (kc->debuglevel > 2)
		fprintf (stderr, "reading packet of %d bytes...\n", len);
	for (num_retry = 5; num_retry; num_retry--) {
		if ((result = read_data(kc, data, len)) < 2)
			return result;
		/*
		 *   Calculate and verify checksum.
		 */
		if (verify_checksum(data + 1, len - 1) >= 0) {
			if (kc->debuglevel > 2)
				fprintf (stderr, "packet checksum correct\n");
			return send_syscode(kc, SYS_CORRECT);
		}
		if (kc->debuglevel > 0)
			fprintf (stderr, "Error: bad checksum from camera\n");
		/*
		 *   Bad checksum.
		 *   Tell camera to resend, unless this
		 *   is the last retry.
		 */
		if (num_retry > 1) {
			if ((result = send_syscode(kc, SYS_ILLEGAL)) < 0)
				return result;
		}
	}
	/*
	 *   Maximum number of retries reached.
	 *   We need to tell the camera that we're giving up.
	 */
	if ((result = send_syscode(kc, SYS_CANCEL)) < 0)
		return result;
	kc->syscode = SYS_ILLEGAL;
	return -2;
}

/*
 *   Generic function to send arbitrary data to the camera.
 *   Partial writes are handled correctly.
 *
 *   Returns 0 on success and -1 on I/O error (errno is set).
 */

static int
send_data (kodacon *kc, byte *data, int len)
{
	int written, result;

	if (kc->debuglevel > 2)
		fprintf (stderr, "writing %d bytes...\n", len);
	for (written = 0; written < len;) {
		while ((result =
		    write(kc->fd_out, data + written, len - written)) < 0
		    && errno == EINTR)
			;
		if (result < 0) {
			if (kc->debuglevel > 0)
				fprintf (stderr, "Error: %s\n",
				    strerror(errno));
			return -1;
		}
		written += result;
	}
	if (kc->debuglevel > 1)
		fprintf (stderr, "wrote %d bytes\n", len);
	return 0;
}

/*
 *   Send a single data packet to the camera and wait for the
 *   camera to tell us that it was received correctly.
 *
 *   This function can be used for sending both auxiliary data
 *   packets and file data packets, although the latter is not
 *   used here yet.  This function will care for the checksum.
 *
 *   Return values:
 *   0    Success.
 *   -1   I/O error, errno is set.
 *   -2   Protocol or camera error, kc->syscode is set.
 */

static int
send_packet (kodacon *kc, byte *data, int len)
{
	int result, num_retry;

	if (kc->debuglevel > 2)
		fprintf (stderr, "writing packet of %d bytes...\n", len);
	create_checksum (data + 1, len - 1);
	for (num_retry = 5; num_retry; num_retry--) {
		if (send_data(kc, data, len) < 0)
			return -1;
		/*
		 *   Wait for the camera to say that the
		 *   packet was received correctly.
		 */
		if ((result = check_syscode(kc, SYS_CORRECT, SYS_ILLEGAL)) < 0)
			return result;
		if (result == SYS_CORRECT) {
			if (kc->debuglevel > 2)
				fprintf (stderr, "packet checksum correct\n");
			return 0;
		}
		if (kc->debuglevel > 0)
			fprintf (stderr, "Error: bad checksum to camera\n");
	}
	/*
	 *   Maximum number of retries reached.
	 *   We need to tell the camera that we're giving up.
	 */
	data[0] = CTRL_CANCEL;
	if (send_data(kc, data, len) < 0)
		return -1;
	kc->syscode = result;
	return -2;
}

/*
 *   Send a command to the camera and handle any system codes
 *   and data transfers, if necessary.  Parameters should
 *   already be written to kc->cmd[2...5].
 *
 *   Return values:
 *   0    Success.
 *   -1   I/O error, errno is set.
 *   -2   Protocol or camera error, kc->syscode is set.
 */

static int
send_cmd (kodacon *kc, int cmd)
{
	int result, flags, pktsize;

	kc->cmdidx = cmd;
	kc->cmd[0] = cmdtab[cmd].code;
	flags = cmdtab[cmd].flags;
	if (kc->debuglevel > 2)
		fprintf (stderr, "Command #%d, code %02x, flags %04x\n",
		    cmd, kc->cmd[0], flags);
	if (flags & CMD_XFER_OUT) {
		/* This is not supported, so fake an execution error. */
		if (kc->debuglevel > 0)
			fprintf (stderr,
			    "Error: CMD_XFER_OUT not supported yet\n");
		kc->syscode = SYS_ERROR;
		return -2;
	}
	/*
	 *   Check if it is necessary to open the CompactFlash
	 *   card first.
	 */
	if ((flags & CMD_OPEN) && kc->is_open != CARD_OPEN) {
		byte cmdtmp[8];

		memcpy (cmdtmp, kc->cmd, 8);
		if ((result = open_card(kc)) < 0)
			return result;
		memcpy (kc->cmd, cmdtmp, 8);
		kc->cmdidx = cmd;
	}
	/*
	 *   Send the command to the camera.
	 */
	if (kc->debuglevel > 1)
		fprintf (stderr,
		    "<-- %02x %02x %02x %02x %02x %02x %02x %02x\n",
		    kc->cmd[0], kc->cmd[1], kc->cmd[2], kc->cmd[3], 
		    kc->cmd[4], kc->cmd[5], kc->cmd[6], kc->cmd[7]);
	if (send_data(kc, kc->cmd, 8) < 0)
		return -1;
	/*
	 *   Wait for the camera to ACKnowledge the command.
	 */
	if ((result = check_syscode(kc, SYS_ACK, -1)) < 0)
		return result;
	/*
	 *   If this command requires sending a packet with
	 *   auxiliary data to the camera, then do so.
	 *   It is assumed that kc->auxdata and kc->auxlen
	 *   are set up by the caller.
	 */
	if (flags & CMD_AUXDATA) {
		if ((result = send_packet(kc, kc->auxdata, kc->auxlen)) < 0)
			return result;
	}
	/*
	 *   If this was a simple command (i.e. without
	 *   file data transfer), just wait for the camera
	 *   to report success or failure.
	 */
	if (flags & CMD_SIMPLE)
		return check_syscode(kc, SYS_SUCCESS, -1);
	/*
	 *   For transfer commands, handle the actual
	 *   data transfer.  Currently, only data input
	 *   (i.e. from camera to host) is supported.
	 */
	if ((pktsize = cmdtab[cmd].pktsize) <= 0)
		pktsize = kc->hpbs;
	kc->datalen = 0;
	for (;;) {
		if ((result = read_packet(kc, kc->buffer, pktsize)) < 0)
			return result;
		if (kc->debuglevel > 2)
			fprintf (stderr, "--> pcb: %02x\n", kc->buffer[0]);
		switch (kc->buffer[0]) {
		case SYS_PACKET:
			kc->data =
			    realloc(kc->data, kc->datalen + pktsize - 2);
			memcpy (kc->data + kc->datalen, kc->buffer + 1,
			    pktsize - 2);
			kc->datalen += pktsize - 2;
			break;
		case SYS_SUCCESS:
			return 0;
		case SYS_BUSY:
			continue;
		default:
			kc->syscode = kc->buffer[0];
			return -2;
		}
		if (kc->datasize < 0)
			kc->datasize =
			    (kc->buffer[1] << 8 | kc->buffer[2]) * 20;
		if (!(flags & CMD_COMPLEX) || kc->datalen >= kc->datasize)
			break;
	}
	/*
	 *   Wait for the final SYS_SUCCESS.
	 */
	return check_syscode(kc, SYS_SUCCESS, -1);
}

static int
send_cmd0 (kodacon *kc, int cmd)
{
	kc->cmd[2] = kc->cmd[3] = kc->cmd[4] = kc->cmd[5] = 0;
	return send_cmd(kc, cmd);
}

static int
send_cmd8 (kodacon *kc, int cmd, byte par)
{
	kc->cmd[2] = par;
	kc->cmd[3] = kc->cmd[4] = kc->cmd[5] = 0;
	return send_cmd(kc, cmd);
}

static int
send_cmd8c (kodacon *kc, int cmd, byte par)
{
	kc->cmd[4] = par;
	kc->cmd[2] = kc->cmd[3] = kc->cmd[5] = 0;
	return send_cmd(kc, cmd);
}

static int
send_cmd16 (kodacon *kc, int cmd, int par)
{
	if (par < 0)
		par += 65536;
	kc->cmd[2] = par >> 8;
	kc->cmd[3] = par & 0xff;
	kc->cmd[4] = kc->cmd[5] = 0;
	return send_cmd(kc, cmd);
}

static int
send_cmd32 (kodacon *kc, int cmd, u_int32_t par)
{
	kc->cmd[2] = par >> 24;
	kc->cmd[3] = par >> 16 & 0xff;
	kc->cmd[4] = par >> 8 & 0xff;
	kc->cmd[5] = par & 0xff;
	return send_cmd(kc, cmd);
}

static void
make_filepacket (kodacon *kc, char *name)
{
	char *src, *dst;
	int i;

	memset (kc->auxdata, 0, 60);
	kc->auxdata[0] = CTRL_FINAL;
	strcpy ((char *) kc->auxdata + 1, "\\PCCARD");
	src = name;
	dst = (char *) kc->auxdata + 8;
	for (i = 0; src[i] && i < 41; i++)
		if (src[i] == '/')
			dst[i] = '\\';
		else
			dst[i] = src[i];
	memset (kc->auxdata + 49, 0xff, 8);
	kc->auxlen = 60;
}

/*
 *   Public functions.  The protoypes for these can be found
 *   in the proto.h file.  They commonly return the following
 *   values, unless otherwise noticed:
 *   0    Success.
 *   -1   I/O error, errno is set.
 *   -2   Protocol or camera error, kc->syscode is set.
 */

int
set_hpbs (kodacon *kc, int new_hpbs)
{
	int result;

	if ((result = send_cmd16(kc, SET_HPBS, new_hpbs)) >= 0) {
		if (kc->debuglevel > 0)
			fprintf (stderr, "set_hpbs() successful\n");
		if (kc->buffer)
			free (kc->buffer);
		kc->buffer = malloc(new_hpbs);
		kc->hpbs = new_hpbs;
	}
	return result;
}

int
set_resolution (kodacon *kc, int rescode)
{
	/*
	 *   0:   640 x 480
	 *   1:  1280 x 960 
	 */
	return send_cmd8(kc, SET_RESOLUTION, rescode);
}

int
set_awb (kodacon *kc, int awbcode)
{
	/*
	 *   0:  auto
	 *   1:  fluorescent
	 *   2:  tungsten
	 *   3:  daylight / flash
	 */
	return send_cmd8(kc, SET_AWB, awbcode);
}

int
set_exposurelock (kodacon *kc, int onoff)
{
	return send_cmd8(kc, SET_EXPOSURELOCK, onoff);
}

int
set_sharpness (kodacon *kc, int sharpcode)
{
	/*
	 *     0:  normal
	 *     1:  sharpen
	 *   255:  soften
	 */
	return send_cmd8(kc, SET_SHARPNESS, sharpcode);
}

int
set_timestamp (kodacon *kc, int stampcode)
{
	/*
	 *   0:  off
	 *   1:  YYYY MM DD      4:  YYYY MM DD hh:mm
	 *   2:  DD MM YYYY      5:  DD MM YYYY hh:mm
	 *   3:  MM DD YYYY      6:  MM DD YYYY hh:mm
	 */
	return send_cmd8(kc, SET_TIMESTAMP, stampcode);
}

int
set_effect (kodacon *kc, int fxcode)
{
	/*
	 *   0x00:  off
	 *   0x01:  grayscale
	 *   0x02:  sepia
	 *   0x04:  document
	 *   0x08:  border
	 */
	return send_cmd8(kc, SET_EFFECT, fxcode);
}

int
set_borderfile (kodacon *kc, char *name)
{
	make_filepacket (kc, name);
	return send_cmd0(kc, SET_BORDERFILE);
}

int
get_lastreviewed (kodacon *kc)
{
	return send_cmd0(kc, GET_LASTREVIEWED);
}

int
get_lasttaken (kodacon *kc)
{
	return send_cmd0(kc, GET_LASTTAKEN);
}

int
set_imagequality (kodacon *kc, int qualitycode)
{
	/*
	 *   3:  good
	 *   2:  better
	 *   1:  best
	 */
	return send_cmd8(kc, SET_IMAGEQUALITY, qualitycode);
}

int
set_flashmode (kodacon *kc, int flashcode)
{
	/*
	 *   0:  auto
	 *   1:  fill
	 *   2:  off
	 *   3:  auto + red-eye reduction
	 */
	return send_cmd8(kc, SET_FLASHMODE, flashcode);
}

int
set_focusmode (kodacon *kc, int focuscode)
{
	/*
	 *   0x02:  auto / spot
	 *   0x22:  close-up
	 *   0x32:  infinity
	 */
	return send_cmd8(kc, SET_FOCUSMODE, focuscode);
}

int
set_shutterdelay (kodacon *kc, int onoff)
{
	return send_cmd8(kc, SET_SHUTTERDELAY, onoff);
}

int
set_time (kodacon *kc, int year, int month, int day,
	  int hour, int min, int sec, int msec)
{
	memset (kc->auxdata, 0, 60);
	kc->auxlen = 60;
	kc->auxdata[0] = year >> 8;
	kc->auxdata[1] = year & 0xff;
	kc->auxdata[2] = month;
	kc->auxdata[3] = day;
	kc->auxdata[4] = hour;
	kc->auxdata[5] = min;
	kc->auxdata[6] = sec;
	kc->auxdata[7] = (msec + 5) / 10;
	return send_cmd0(kc, SET_TIME);
}

int
set_zoom (kodacon *kc, int zoom)
{
	/*
	 *   100 - 300 means 1.00 - 3.00 (optical zoom only),
	 *   301 - 600 means 3.01 - 6.00 (uses optical + digital zoom).
	 */
	return send_cmd32(kc, SET_ZOOM, zoom);
}

int
set_autoexposure (kodacon *kc, int aecode)
{
	/*
	 *   0:  auto
	 *   1:  center-weighted
	 */
	return send_cmd8(kc, SET_SHUTTERDELAY, aecode);
}

int
take_picture (kodacon *kc)
{
	return send_cmd0(kc, TAKE_PICTURE);
}

int
check_battery (kodacon *kc)
{
	/*
	 *   IMPORTANT:
	 *   ==========
	 *   When you use this command, the camera draws
	 *   current from the batteries for a short time
	 *   to check their state.  That means that this
	 *   command consumes a certain amount of battery
	 *   power, and you will drain the batteries if
	 *   it is used often!
	 *
	 *   After using this command, the battery status byte
	 *   in the camera status table is updated.  Use the
	 *   get_camerastatus() function to read it.
	 */
	return send_cmd0(kc, CHECK_BATTERY);
}

int read_camerastatus (kodacon *kc)
{
	return send_cmd0(kc, READ_CAMERASTATUS);
}

int
set_manualexposure (kodacon *kc, int exposure)
{
	/*
	 *   -200 - 200 means -2.00 - 2.00.
	 */
	return send_cmd16(kc, SET_MANUALEXPOSURE, exposure);
}

int
reset_camera (kodacon *kc)
{
	/*
	 *   Resets camera to system defaults.
	 */
	return send_cmd0(kc, RESET_CAMERA);
}

int
execute_diagnostic (kodacon *kc, int diagcode)
{
	/*
	 *   0x0100:  switch monitor
	 *   0x0200:  motor test
	 *   0x0300:  CdS sensor test
	 */
	return send_cmd16(kc, EXECUTE_DIAGNOSTIC, diagcode);
}

int
read_pictureinfo (kodacon *kc, char *name)
{
	make_filepacket (kc, name);
	return send_cmd0(kc, READ_PICTUREINFO);
}

int
read_thumbnail (kodacon *kc, char *name, unsigned int size)
{
	make_filepacket (kc, name);
	kc->datasize = size;
	return send_cmd8c(kc, READ_THUMBNAIL, 2);
}

int
format_compactflash (kodacon *kc, char *volumeID)
{
	memset (kc->auxdata, 0, 60);
	kc->auxlen = 60;
	strncpy ((char *) kc->auxdata, volumeID, 11);
	return send_cmd0(kc, FORMAT_COMPACTFLASH);
}

int
open_card (kodacon *kc)
{
	int result;

	/*
	 *   If the current card state is unknown, then close
	 *   it first, just to be sure.  This is a no-op and
	 *   will not cause an error if it has not been opened.
	 *   (But note that trying to open the card _will_
	 *   cause an error if it has already been opened!)
	 */
	if (kc->is_open != CARD_CLOSED)
		if ((result = close_card(kc)) < 0)
			return result;
	if ((result = send_cmd0(kc, OPEN_CARD)) < 0) {
		kc->is_open = CARD_UNKNOWN;
		return result;
	}
	if (kc->debuglevel > 0)
		fprintf (stderr, "card_open() successful\n");
	kc->is_open = CARD_OPEN;
	return 0;
}

int
close_card (kodacon *kc)
{
	int result;

	if ((result = send_cmd0(kc, CLOSE_CARD)) < 0) {
		kc->is_open = CARD_UNKNOWN;
		return result;
	}
	if (kc->debuglevel > 0)
		fprintf (stderr, "card_close() successful\n");
	kc->is_open = CARD_CLOSED;
	return 0;
}

int
get_cardstatus (kodacon *kc)
{
	return send_cmd0(kc, GET_CARDSTATUS);
}

int
read_directory (kodacon *kc, char *name)
{
	make_filepacket (kc, name);
	kc->datasize = -1;
	return send_cmd0(kc, READ_DIRECTORY);
}

int
read_file (kodacon *kc, char *name, unsigned int size)
{
	make_filepacket (kc, name);
	kc->datasize = size;
	return send_cmd0(kc, READ_FILE);
}

int
write_file (kodacon *kc, char *name, unsigned int size)
{
	/*
	 *   This is not supported yet in send_cmd()!
	 */
	make_filepacket (kc, name);
	kc->datasize = size;
	return send_cmd0(kc, WRITE_FILE);
}

int
delete_file (kodacon *kc, char *name)
{
	make_filepacket (kc, name);
	return send_cmd0(kc, DELETE_FILE);
}

int
write_cameraid (kodacon *kc, char *cameraID)
{
	memset (kc->auxdata, 0, 60);
	kc->auxlen = 60;
	strncpy ((char *) kc->auxdata, cameraID, 32);
	return send_cmd0(kc, WRITE_CAMERAID);
}

int
set_protect (kodacon *kc, char *name, int onoff)
{
	make_filepacket (kc, name);
	return send_cmd8(kc, SET_PROTECT, onoff);
}

int
proto_init (kodacon *kc, int fd_out, int fd_in, int verbosity, int debuglevel)
{
	memset (kc, 0, sizeof(kodacon));
	kc->fd_out = fd_out;
	kc->fd_in = fd_in;
	kc->debuglevel = debuglevel;
	kc->verbosity = verbosity;
	kc->hpbs = 0x0202;
	kc->buffer = malloc(kc->hpbs);
	kc->is_open = CARD_UNKNOWN;
	kc->cmd[7] = 0x1a;
	kc->auxdata = malloc(60);
	return 0;
}

int
proto_done (kodacon *kc)
{
	if (kc->is_open != CARD_CLOSED) {
		int result;

		if ((result = close_card(kc)) < 0)
			return result;
	}
	if (kc->buffer)
		free (kc->buffer);
	if (kc->auxdata)
		free (kc->auxdata);
	if (kc->data)
		free (kc->data);
	return 0;
}

char *
strsyscode (int syscode)
{
	static char *str = NULL;

	switch (syscode) {
	case SYS_SUCCESS:	return "command completed successfully";
	case SYS_ACK:		return "command accepted";
	case SYS_CORRECT:	return "packet transferred successfully";
	case SYS_NAK:		return "command not accepted";
	case SYS_ERROR:		return "command failed";
	case SYS_ILLEGAL:	return "packet transfer failed";
	case SYS_CANCEL:	return "packet transfer cancelled";
	case SYS_BUSY:		return "camera busy";
	default:
		if (str)
			free (str);
		asprintf (&str, "unknown system code %02xh", syscode);
		return str;
	}
}

int
big2ss (byte *big)
{
	int ss;

	ss = (unsigned int) big[0] << 8 | (unsigned int) big[1];
	if (ss > 32767)
		ss -= 65536;
	return ss;
}

unsigned int
big2us (byte *big)
{
	return (unsigned int) big[0] << 8 | (unsigned int) big[1];
}

unsigned long
big2ul (byte *big)
{
	return (unsigned long) big[0] << 24 | (unsigned long) big[1] << 16
	     | (unsigned long) big[2] << 8  | (unsigned long) big[3];
}

/* EOF */
