/*
 * proto.h
 *
 * Copyright (c) 1999  Oliver Fromme  <olli@fromme.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *     @(#)$Id: proto.h,v 1.9 1999/08/07 16:59:03 olli Exp $
 */

#ifndef _OPHOTO_PROTO_H_
#define _OPHOTO_PROTO_H_

static const char cvsid_proto_h[]
    = "@(#)$Id: proto.h,v 1.9 1999/08/07 16:59:03 olli Exp $";

typedef unsigned char byte;

/* struct for connection status */

typedef struct {
	int fd_out, fd_in;	/* file descriptors for output and input */
	int hpbs;		/* host packet buffer size (def. 0x0202) */
	byte cmdidx;		/* index of last command code */
	byte cmd[8];		/* command sent to camera */
	byte *auxdata;		/* auxiliary data packet for command... */
	int auxlen;		/* ... and its length (60 bytes max.) */
	byte syscode;		/* system code in case of error */
	byte *buffer;		/* packet transfer buffer (hpbs bytes) */
	byte *data;		/* file data for transfer... */
	int datalen;		/* ... and its current length */
	int datasize;		/* ... and its expected length */
	int is_open;		/* CompactFlash open state (see below) */
	int debuglevel;		/* debug level */
	int verbosity;		/* verbosity level */
} kodacon;

/* struct for command information */

typedef struct {
	byte code;	/* command code */
	int pktsize;	/* packet size incl. PCB (0 = none, -1 = use hpbs) */
	int flags;	/* command flags, see below */
} kodacmd;

/*
 *   command flags
 *
 *   In thise code we differentiate three types of commands:
 *
 *    - "Simple" commands which don't involve transfer of user
 *      data (CMD_SIMPLE).
 *
 *    - Generic input / output commands which use a single
 *      packet to transfer user data in either direction
 *      (CMD_XFER_OUT, CMD_XFER_IN).
 *
 *    - "Complex" commands which use an arbitrary number of
 *      packets to transfer user data in either direction.
 *      (CMD_COMPLEX | CMD_XFER_{IN,OUT}).
 *
 *   Any of these command types can require transferring a
 *   single packet containing auxiliary data (for command
 *   parameters) to the camera, which is immediately following
 *   the command.  This is indicated by CMD_AUXDATA.
 *   For the purpose of this implementation, auxiliary packets
 *   are never larger than 6 bytes (incl. PCB and checksum).
 *   If a command requires more data, we use CMD_XFER_OUT
 *   with a single data packet instead.  (There are no
 *   commands which require more than 60 bytes of auxiliary
 *   data _and_ some user data at the same time.)
 *
 *   Finally, certain commands are only allowed if the memory
 *   card has been opened before.  This is indicated by the
 *   CMD_OPEN flag.
 */

#define CMD_SIMPLE	0x0001
#define CMD_XFER_IN	0x0002
#define CMD_XFER_OUT	0x0004
#define CMD_COMPLEX	0x0008
#define CMD_AUXDATA	0x0010
#define CMD_OPEN	0x0020

/*
 *   system codes
 *
 *   SYS_SUCCESS is sent by the camera after every command
 *   that has been completed successfully.  After this,
 *   a delay of 50ms is required before the next command
 *   can be sent.  (This is the opposite of SYS_ERROR.)
 *
 *   0x01 (``SYS_PACKET'') is actually not a real system code,
 *   but a packet control byte (PCB) sent from the camera to
 *   the host, indicating that a data packet follows.  However,
 *   we use it in the context of system codes.  This is really
 *   only a matter of interpretation.
 *
 *   SYS_ACK (``acknowledge'') is sent to the host after the
 *   command (8 bytes) has been received and accepted by the
 *   camera, but before data transfer happens (if any).
 *   (This is the opposite of SYS_NAK.)  If the command does
 *   not involve any data transfer, SYS_ACK is immediately
 *   followed by SYS_SUCCESS (or SYS_ERROR in case of an
 *   error).
 *
 *   SYS_CORRECT is sent in either direction (camera --> host
 *   or host --> camera, as appropriate) after every packet
 *   of user data to indicate that the transfer was successful.
 *   (This is the opposite of SYS_ILLEGAL.)
 *
 *   SYS_NAK (``no acknowledge'') is sent to the host instead
 *   of SYS_ACK if the camera does not accept a command (for
 *   example if the command is not supported, or if it is
 *   malformed in some way).
 *
 *   SYS_ERROR is sent by the camera instead of SYS_SUCCESS if
 *   a command was not completed successfully.  Examples for
 *   this are: file not found, parameter out of range, no
 *   memory card present, etc.
 *
 *   SYS_ILLEGAL is sent in either direction (camera --> host
 *   or host --> camera, as appropriate) if a user data packet
 *   was not transferred successfully (i.e. checksum error or
 *   illegal packet control byte).  There are two cases:
 *    - Data transfer from camera to host:
 *      After the host has sent SYS_ILLEGAL to the camera, it
 *      will immediately try again and repeat the transfer.
 *      If you want to abort, do not send SYS_ILLEGAL, but
 *      SYS_CANCEL.
 *    - Data transfer from host to camera:
 *      After the camera has sent SYS_ILLEGAL, it expects that
 *      the host will immediately try angain and repeat the
 *      transfer.  If you want to abort, the packet control
 *      byte (PCB) of the packet should be set to CTRL_CANCEL
 *      (the camera will then ignore the rest of the packet).
 *
 *   SYS_CANCEL can be sent to the camera in response to a
 *   packet received from the camera (instead of SYS_ILLEGAL
 *   or SYS_CORRECT), to indicate that the transfer should
 *   be aborted immediately.  (To abort a transfer in the
 *   other direction, send a packet with the PCB byte set to
 *   CTRL_CANCEL).
 *
 *   SYS_BUSY can be sent by the camera any time when any
 *   other system code or a user data packet can be sent.
 *   It indicates that the camera needs a bit more time
 *   (> 2s) to continue the request.  After that, it simply
 *   continues.  Most of the time we can simply ignore
 *   SYS_BUSY and wait for the next system code.
 */

#define SYS_SUCCESS	0x00
#define SYS_PACKET	0x01
#define SYS_ACK		0xd1
#define SYS_CORRECT	0xd2
#define SYS_NAK		0xe1
#define SYS_ERROR	0xe2
#define SYS_ILLEGAL	0xe3
#define SYS_CANCEL	0xe4
#define SYS_BUSY	0xf0

/*
 *   packet control bytes (PCB)
 *
 *   Every packet of user data contains such a PCB in the
 *   very first byte, and a checksum in the very last byte.
 *
 *   Packets sent from the camera to the host always use
 *   CTRL_CAMERA.
 *
 *   Packets sent from the host to the camera can use these:
 *    - CTRL_NORMAL indicates a normal packet with user data
 *      which is not the last packet of the current transfer.
 *    - CTRL_FINAL are set in the last (or only) packet of
 *      a transfer of user data.  Note that auxiliary data
 *      packets (see the command flag CMD_AUXDATA above) are
 *      considered to be a single transfer, so they always
 *      use CTRL_FINAL in the PCB.
 *    - CTRL_CANCEL indicates that the host wishes to abort
 *      the transfer, and that no more packets will follow.
 *      Note that you still have to send a full-size packet
 *      when you use this PCB, but the camera will ignore
 *      the contents.
 */

#define CTRL_NORMAL	0x00
#define CTRL_CAMERA	0x01
#define CTRL_FINAL	0x80
#define CTRL_CANCEL	0xff

/*
 *   CompactFlash open state
 *
 *   Some commands require the CompactFlash card to be
 *   "opened" before (see the command flag CMD_OPEN above).
 *   Note that it is an error to try to open the card if
 *   it has already been opened before, but it is not an
 *   error to close the card if it has not been opened
 *   (this is a no-op).  Therefore we treat the UNKNOWN
 *   state the same as CARD_OPEN and close it if necessary,
 *   just to be sure (it doesn't hurt in any case).
 */

#define CARD_OPEN	1
#define CARD_CLOSED	0
#define CARD_UNKNOWN	-1

/*
 *   Public function prototypes.  They commonly return the
 *   following values, unless otherwise noticed:
 *   0    Success.
 *   -1   I/O error, errno is set.
 *   -2   Protocol or camera error, kc->syscode is set.
 */

int set_hpbs		(kodacon *kc, int new_hpbs);
int set_resolution	(kodacon *kc, int rescode);
int set_awb		(kodacon *kc, int awbcode);
int set_exposurelock	(kodacon *kc, int onoff);
int set_sharpness	(kodacon *kc, int sharpcode);
int set_timestamp	(kodacon *kc, int stampcode);
int set_effect		(kodacon *kc, int fxcode);
int set_borderfile	(kodacon *kc, char *name);
int get_lastreviewed	(kodacon *kc);
int get_lasttaken	(kodacon *kc);
int set_imagequality	(kodacon *kc, int qualitycode);
int set_flashmode	(kodacon *kc, int flashcode);
int set_focusmode	(kodacon *kc, int focuscode);
int set_shutterdelay	(kodacon *kc, int onoff);
int set_time		(kodacon *kc, int year, int month, int day,
			 int hour, int min, int sec, int msec);
int set_zoom		(kodacon *kc, int zoom);
int set_autoexposure	(kodacon *kc, int aecode);
int take_picture	(kodacon *kc);
int check_battery	(kodacon *kc);	/* READ COMMENT IN proto.c!!! */
int read_camerastatus	(kodacon *kc);
int set_manualexposure	(kodacon *kc, int exposure);
int reset_camera	(kodacon *kc);
int execute_diagnostic	(kodacon *kc, int diagcode);
int read_pictureinfo	(kodacon *kc, char *name);
int read_thumbnail	(kodacon *kc, char *name, unsigned int size);
int format_compactflash	(kodacon *kc, char *volumeID);
int open_card		(kodacon *kc);
int close_card		(kodacon *kc);
int get_cardstatus	(kodacon *kc);
int read_directory	(kodacon *kc, char *name);
int read_file		(kodacon *kc, char *name, unsigned int size);
int delete_file		(kodacon *kc, char *name);
int write_cameraid	(kodacon *kc, char *cameraID);
int set_protect		(kodacon *kc, char *name, int onoff);

int proto_init (kodacon *kc, int fd_out, int fd_in,
		int verbosity, int debuglevel);
int proto_done (kodacon *kc);

char *strsyscode (int syscode);
int           big2ss (byte *big);
unsigned int  big2us (byte *big);
unsigned long big2ul (byte *big);

#endif /* #ifndef _OPHOTO_PROTO_H_ */

/* EOF */
